use std::{
    collections::{BTreeMap, HashMap},
    ptr::NonNull,
};

use crate::{
    ast::{Node as AstNode, Tag},
    parser::Tree,
};

#[derive(Debug)]
pub struct SymbolRecord {
    name: String,
    decl: AstNode,
}

impl SymbolRecord {
    pub fn node(&self) -> AstNode {
        self.decl
    }

    pub fn name(&self) -> &str {
        &self.name
    }
}

#[allow(unused)]
pub struct SymbolPath(Vec<Option<AstNode>>, String);

impl SymbolPath {
    pub fn mangle(&self, tree: &Tree) -> String {
        use core::fmt::Write;
        let mut buf = String::new();

        for node in self.0.iter().skip(1).rev() {
            match tree.nodes.get_node(node.unwrap()) {
                Tag::VarDecl { name, .. } => {
                    _ = write!(&mut buf, "V{}::", tree.get_ident_str(*name).unwrap());
                }
                Tag::GlobalDecl { name, .. } => {
                    _ = write!(&mut buf, "G{}::", tree.get_ident_str(*name).unwrap());
                }
                Tag::FunctionProto { name, .. } => {
                    _ = write!(&mut buf, "F{}::", tree.get_ident_str(*name).unwrap());
                }
                _ => {}
            }
        }
        _ = write!(&mut buf, "{}", self.1);
        buf
    }
}

pub enum SymbolKind {
    Var, // or Let
    Orderless,
}

#[derive(Debug, Default)]
struct InnerSymbolTable {
    ordered_identifiers: Vec<SymbolRecord>,
    orderless_identifiers: HashMap<String, SymbolRecord>,
    children: BTreeMap<AstNode, NonNull<InnerSymbolTable>>,
    scope: Option<AstNode>,
    parent: Option<NonNull<InnerSymbolTable>>,
}
impl InnerSymbolTable {
    fn new() -> NonNull<InnerSymbolTable> {
        Self::new_with(Self::new_inner)
    }

    fn new_with<G>(gen: G) -> NonNull<InnerSymbolTable>
    where
        G: FnOnce() -> Self,
    {
        NonNull::new(Box::leak(Box::new(gen())) as *mut _).unwrap()
    }

    fn new_inner() -> InnerSymbolTable {
        Self {
            parent: None,
            ordered_identifiers: Vec::new(),
            orderless_identifiers: HashMap::new(),
            children: BTreeMap::new(),
            scope: None,
        }
    }

    fn make_child(&self, scope: AstNode) -> NonNull<InnerSymbolTable> {
        Self::new_with(|| Self {
            parent: NonNull::new(self.as_ptr()),
            ordered_identifiers: Vec::new(),
            orderless_identifiers: HashMap::new(),
            children: BTreeMap::new(),
            scope: Some(scope),
        })
    }

    fn parent(&self) -> Option<NonNull<InnerSymbolTable>> {
        self.parent
    }

    fn parent_ref(&self) -> Option<&InnerSymbolTable> {
        unsafe { self.parent.map(|p| p.as_ref()) }
    }

    fn parent_mut(&mut self) -> Option<&mut InnerSymbolTable> {
        unsafe { self.parent.map(|mut p| p.as_mut()) }
    }

    fn as_ptr(&self) -> *mut Self {
        self as *const _ as *mut _
    }

    fn root(&self) -> NonNull<InnerSymbolTable> {
        self.parent()
            .map(|p| unsafe { p.as_ref().root() })
            .unwrap_or(NonNull::new(self.as_ptr()).unwrap())
    }
}

impl Drop for InnerSymbolTable {
    fn drop(&mut self) {
        for child in self.children.values() {
            unsafe {
                _ = Box::from_raw(child.as_ptr());
            }
        }
    }
}

impl InnerSymbolTable {
    fn insert_symbol(&mut self, name: &str, node: AstNode, kind: SymbolKind) -> &SymbolRecord {
        match kind {
            SymbolKind::Var => {
                self.ordered_identifiers.push(SymbolRecord {
                    name: name.to_owned(),
                    decl: node,
                });
                self.ordered_identifiers.last().unwrap()
            }
            _ => self.insert_orderless_symbol(name, node),
        }
    }

    fn insert_orderless_symbol(&mut self, name: &str, node: AstNode) -> &SymbolRecord {
        self.orderless_identifiers.insert(
            name.to_owned(),
            SymbolRecord {
                name: name.to_owned(),
                decl: node,
            },
        );
        self.orderless_identifiers.get(name).unwrap()
    }

    fn find_symbol_or_insert_with<'a, F>(&'a mut self, name: &str, cb: F) -> &'a SymbolRecord
    where
        F: FnOnce() -> (AstNode, SymbolKind),
    {
        let this = self as *mut Self;
        if let Some(record) = unsafe { &*this }.find_any_symbol(name) {
            record
        } else {
            let (node, kind) = cb();
            self.insert_symbol(name, node, kind)
        }
    }

    fn find_symbol_by_decl(&self, decl: AstNode) -> Option<&SymbolRecord> {
        self.ordered_identifiers
            .iter()
            .find(|r| r.decl == decl)
            .or_else(|| {
                self.orderless_identifiers
                    .iter()
                    .find(|(_, v)| v.decl == decl)
                    .map(|(_, v)| v)
            })
            .or_else(|| self.parent_ref().and_then(|p| p.find_symbol_by_decl(decl)))
    }

    fn find_any_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.ordered_identifiers
            .iter()
            .find(|r| r.name.as_str() == name)
            .or_else(|| self.orderless_identifiers.get(name))
            .or_else(|| self.parent_ref().and_then(|p| p.find_any_symbol(name)))
    }

    fn find_ordered_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.ordered_identifiers
            .iter()
            .find(|r| r.name.as_str() == name)
            .or_else(|| self.parent_ref().and_then(|p| p.find_ordered_symbol(name)))
    }

    fn find_orderless_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.orderless_identifiers.get(name).or_else(|| {
            self.parent_ref()
                .and_then(|p| p.find_orderless_symbol(name))
        })
    }

    fn extend_orderless<I>(&mut self, iter: I)
    where
        I: IntoIterator<Item = (String, SymbolRecord)>,
    {
        self.orderless_identifiers.extend(iter)
    }

    fn extract_orderless_if<F>(
        &mut self,
        pred: F,
    ) -> std::collections::hash_map::ExtractIf<String, SymbolRecord, F>
    where
        F: FnMut(&String, &mut SymbolRecord) -> bool,
    {
        self.orderless_identifiers.extract_if(pred)
    }
}

#[derive(Debug)]
pub struct SymbolTableWrapper {
    current: NonNull<InnerSymbolTable>,
}

impl Drop for SymbolTableWrapper {
    fn drop(&mut self) {
        unsafe {
            _ = Box::from_raw(self.current.as_ref().root().as_ptr());
        }
    }
}

impl SymbolTableWrapper {
    pub fn new() -> SymbolTableWrapper {
        Self {
            current: InnerSymbolTable::new(),
        }
    }

    fn current(&self) -> &InnerSymbolTable {
        unsafe { self.current.as_ref() }
    }

    fn current_mut(&mut self) -> &mut InnerSymbolTable {
        unsafe { self.current.as_mut() }
    }

    #[allow(dead_code)]
    fn root_ref(&self) -> &InnerSymbolTable {
        unsafe { self.current().root().as_ref() }
    }

    fn root_mut(&mut self) -> &mut InnerSymbolTable {
        unsafe { self.current_mut().root().as_mut() }
    }

    #[allow(dead_code)]
    fn parent_ref(&self) -> Option<&InnerSymbolTable> {
        self.current().parent_ref()
    }

    #[allow(dead_code)]
    fn parent_mut(&mut self) -> Option<&mut InnerSymbolTable> {
        self.current_mut().parent_mut()
    }

    pub fn into_child(&mut self, scope: AstNode) {
        let child = if let Some(child) = self.current().children.get(&scope) {
            *child
        } else {
            let child = self.current().make_child(scope);
            self.current_mut().children.insert(scope, child);
            child
        };
        self.current = child;
    }

    pub fn into_parent(&mut self) {
        if let Some(parent) = self.current().parent() {
            self.current = parent;
        }
    }
}

impl SymbolTableWrapper {
    pub fn insert_symbol(&mut self, name: &str, node: AstNode, kind: SymbolKind) -> &SymbolRecord {
        self.current_mut().insert_symbol(name, node, kind)
    }

    pub fn find_root_symbol(&mut self, name: &str) -> Option<&SymbolRecord> {
        self.root_mut().find_orderless_symbol(name)
    }

    pub fn insert_root_symbol(&mut self, name: &str, node: AstNode) -> &SymbolRecord {
        self.root_mut().insert_orderless_symbol(name, node)
    }

    pub fn insert_orderless_symbol(&mut self, name: &str, node: AstNode) -> &SymbolRecord {
        self.current_mut().insert_orderless_symbol(name, node)
    }

    pub fn find_symbol_or_insert_with<'a, F>(&'a mut self, name: &str, cb: F) -> &'a SymbolRecord
    where
        F: FnOnce() -> (AstNode, SymbolKind),
    {
        self.current_mut().find_symbol_or_insert_with(name, cb)
    }

    pub fn find_symbol_by_decl(&self, decl: AstNode) -> Option<&SymbolRecord> {
        self.current().find_symbol_by_decl(decl)
    }

    pub fn find_any_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.current().find_any_symbol(name)
    }

    pub fn find_ordered_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.current().find_ordered_symbol(name)
    }

    pub fn find_orderless_symbol(&self, name: &str) -> Option<&SymbolRecord> {
        self.current().find_orderless_symbol(name)
    }

    pub fn symbol_path(&self, decl: AstNode) -> Option<SymbolPath> {
        let mut table = self.current();
        loop {
            if let Some(record) = table
                .ordered_identifiers
                .iter()
                .find(|r| r.node() == decl)
                .or_else(|| {
                    table
                        .orderless_identifiers
                        .iter()
                        .find(|(_, v)| v.decl == decl)
                        .map(|(_, v)| v)
                })
            {
                let mut path = Vec::new();
                while let Some(parent) = table.parent_ref() {
                    path.push(parent.scope);
                }

                return Some(SymbolPath(path, record.name.clone()));
            };

            let Some(parent) = table.parent_ref() else {
                break;
            };
            table = parent;
        }

        None
    }

    pub fn extend_orderless<I>(&mut self, iter: I)
    where
        I: IntoIterator<Item = (String, SymbolRecord)>,
    {
        self.current_mut().extend_orderless(iter)
    }

    pub fn extract_orderless_if<F>(
        &mut self,
        pred: F,
    ) -> std::collections::hash_map::ExtractIf<String, SymbolRecord, F>
    where
        F: FnMut(&String, &mut SymbolRecord) -> bool,
    {
        self.current_mut().extract_orderless_if(pred)
    }
}

pub type SymbolTable = SymbolTableWrapper;
